from numpy import ndarray
import pyedflib
import math
from typing import Any, Callable,Union
def get_all_signals(edf:pyedflib.EdfReader):
    """
    get all signals
    """
    return [edf.readSignal(idx) for idx,_ in enumerate(edf.getSignalLabels())]

def get_fs(edf:pyedflib.EdfReader,chn=0):
    """
    get sample rate
    """
    return edf.getSignalHeader(chn)['sample_rate'] # = fs


GetAnnotationType = tuple[str,float,Any,int]
"""
tuple(name,time(s),duration,wave index)
"""


def get_annotations(edf:pyedflib.EdfReader):
    """
    get  annotations
    Returns:
       array of tuple(name,time(s),duration,wave index)
    """
    annotations = edf.readAnnotations()
    fs = get_fs(edf)
    rows:list[GetAnnotationType] = [(name,time,duration,math.floor(time * fs))for time,duration,name in zip(annotations[0],annotations[1],annotations[2])]
    return rows
def get_channels_length(edf:pyedflib.EdfReader):
    """
    get channels length
    """
    return len(edf.getSignalHeaders())
CopiedFuncType = Callable[[pyedflib.EdfReader,pyedflib.EdfWriter,list[ndarray]],Union[list[ndarray],None]]
def copy(redf:pyedflib.EdfReader,copy_path:str,copied_func:Union[CopiedFuncType,None] = None):
    """
    edf file copy
    """
    ch = get_channels_length(redf)
    with pyedflib.EdfWriter(copy_path,ch) as wedf:
        header = redf.getHeader()
        header["birthdate"] = ""
        annos = redf.readAnnotations()
        wedf.setHeader(header)
        wedf.setSignalHeaders(redf.getSignalHeaders())
        all_signals = get_all_signals(redf)
        
        for i,_ in enumerate(annos[0]):
            wedf.writeAnnotation(annos[0][i],annos[1][i],annos[2][i])
        if not (copied_func is None):
            _all_signals = copied_func(redf,wedf,all_signals)
            if not (_all_signals is None):
                all_signals = _all_signals
        wedf.writeSamples(all_signals)
        wedf.close()
    return ch


# %%
import pyedflib
from typing import Any, Optional
from .utilities import edf
import numpy as np
import h5py
from typing import Callable
def split_annotations_edf2hdf(edf_path:str,
    export_path:str,
    is_overwrite:bool = False,
    is_groupby:bool = False,
    filters:list[str] = None,
    before_preprocessing_func:Optional[Callable[[list[np.ndarray]],Any]] = None,
    split_signals_func:Optional[Callable[[Any,edf.GetAnnotationType,Optional[edf.GetAnnotationType]],np.ndarray]] = None,
    after_preprocessing_func:Callable[[np.ndarray,str],np.ndarray] = None,
    end_marker_name:Optional[str] = "__End__"):
    """Split the edf file by annotation and save it in the hdf file.
     Args:
        edf_path : read edf path
        export_path : write hdf path
        is_overwrite : overwrite the edf file
        is_groupby : grouping
        filters : annotation filters
        before_preprocessing_func(function?) : before preprocessing function (split_signals_func must also be used when changing the signal length)
        split_signals_func(function?) : signal split function
        after_preprocessing_func(function[[signals,label],ndarray]?) : Preprocess the signals split by annotations. ndarray : ch × annotation range 
        end_marker_name(str?) : annotation of marker_name end time
    """
    with pyedflib.EdfReader(edf_path) as edf_reader:
        signals = edf.get_all_signals(edf_reader)
        if before_preprocessing_func is not None:
            signals = before_preprocessing_func(signals)
        # simple type check(np.ndarray or list[np.ndarray])
        common_attrs = {}
        if isinstance(signals,np.ndarray) or \
           (isinstance(signals,list) and isinstance(signals[0],np.ndarray) and [type(s) for s in signals].count(type(signals[0])) == len(signals)):
            signals = np.array(signals)
            common_attrs = {
            "presplit_mean":np.mean(signals,axis=1),
            "presplit_std" :np.std(signals,axis=1)
            }
        annotations =  edf.get_annotations(edf_reader)
        split_signals:list[tuple[str,np.ndarray,str,dict]] = [] #[(annotation name,signal,label)]
        for idx,ann in enumerate(annotations):
            next_ann:Optional[edf.GetAnnotationType] = annotations[idx + 1] if idx + 1 < len(annotations) else None
            ann_name,ann_time,_,ann_idx = ann
            if ann_name == end_marker_name:
                continue
            split_ann_name = ann_name.split("_")
            ann_group_name = "_".join(split_ann_name[:-1]) if len(split_ann_name) > 1 else ann_name
            label= split_ann_name[-1] if len(split_ann_name) > 1 else ""
            if split_signals_func is None:
                signals = np.array(signals)
                split_signals.append((ann_group_name,signals[:,ann_idx:(next_ann[3] if next_ann is not None else signals.shape[1])],label,common_attrs))
            else:
                s = split_signals_func(signals,ann,next_ann)
                split_signals.append((ann_group_name,s,label,common_attrs))
        if not(filters is None):
            split_signals =  [ss for ss in split_signals if ss[0] in filters]
        if not(after_preprocessing_func is None):
            split_signals = [(ann_name,after_preprocessing_func(ann_signals,label),label,common_attrs) for ann_name,ann_signals,label,common_attrs in split_signals]
    with h5py.File(export_path, mode='r+' if is_overwrite else 'w') as f:
        ann_group = f.require_group("/annotations")
        if is_groupby:
            for idx,ann in enumerate(split_signals):
                ann_group_name,ann_signals,label,attrs = ann
                local_group = ann_group.require_group("./" + ann_group_name)
                counter = local_group.attrs.get("count")
                counter = 0 if counter is None else counter + 1
                local_group.attrs["count"] = counter
                d = local_group.create_dataset(f"{counter}",ann_signals.shape,data=ann_signals)
                for key in attrs.keys():
                    d.attrs[key] = attrs[key]
                d.attrs["label"] = label

        else:
            for idx,ann in enumerate(split_signals):
                ann_group_name,ann_signals,label,attrs = ann
                d = ann_group.create_dataset(f"{idx}.{ann_group_name}",ann_signals.shape,data=ann_signals)
                for key in attrs.keys():
                    d.attrs[key] = attrs[key]
                d.attrs["label"] = label
# %%



from setuptools import setup
import labedf
with open('README.md', 'r', encoding='utf-8') as fp:
    readme = fp.read()
LONG_DESCRIPTION = readme
LONG_DESCRIPTION_CONTENT_TYPE = 'text/markdown'
setup(
    name="labedf",
    author="sn-10",
    url="https://github.com/s-n-1-0/labedf.py",
    download_url="https://github.com/s-n-1-0/labedf.py",
    version=labedf.__version__,
    description="Merge the lab.js csv file and the edf file.",
    install_requires=[
        "pyedflib>=0.1.28",
        "labcsv>=1.0.3",
        "numpy>=1.22.4",
        "h5py>=3.2.1"
        ],
    packages=["labedf","labedf.utilities"],
    license="MIT",
    keywords= ["labjs","lab.js"],
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown"
)


import os
from numpy import ndarray
import pyedflib
import labcsv
from labcsv import DefaultHeaderName as DHName
from typing import Callable, Optional
from .utilities import edf
def merge_csv2edf(edf_path:str,
                csv_path:str,
                export_path:Optional[str] = None,
                marker_name:str = "Marker",
                sync_marker_name:str = "sync",
                end_marker_name:Optional[str]="__End__",
                end_marker_offset:float = 0,
                label_header_name:str = None,
                preprocessing_func:Optional[Callable[[list[ndarray]],list[ndarray]]] = None):
    """
    Write the lab.js sender names to edf as annotations. (Generate a copy file)
    Args:
        edf_path (str): edf file path
        csv_path (str): csv file path
        export_path (str?): output file path. Defaults to None.(None is <edf_path + "-copy">)
        marker_name(str) : filter sender name(= "sender" value)
        sync_marker_name(str) : "response" value to synchronize files (None = 0 index)
        end_marker_name(str?) : annotation of marker_name end time
        end_marker_offset(float) : marker_name end time offset (seconds)
        label_header_name(str?) : label header name
        preprocessing_func(function) : preprocessing function
    """
    edf_dir_path = os.path.dirname(edf_path)
    edf_filename_path =  os.path.splitext(os.path.basename(edf_path))[0]
    if export_path is None:
        export_path = f"{edf_dir_path}/{edf_filename_path}_copy.edf"
    edf_reader = pyedflib.EdfReader(edf_path)
    rlab_dtype = {label_header_name:str} if not (label_header_name is None) else None
    rlab = labcsv.read_csv(csv_path,dtype=rlab_dtype)

    edf_annos = edf.get_annotations(edf_reader)
    senders,responses,time_ends,time_runs = rlab.get_column_list([DHName.SENDER,DHName.RESPONSE,DHName.TIME_END,DHName.TIME_RUN])
    labels = rlab.get_column_values(label_header_name) if not(label_header_name is None) else [None] * len(responses)
    sync_edf_annos = [ea for ea in edf_annos if ea[0] == sync_marker_name]
    sync_lab_annos_indexes = [i for i,r in enumerate(list(responses)) if r == sync_marker_name ]
    if len(sync_edf_annos) != len(sync_lab_annos_indexes):
        raise Exception("Number of sync_marker_name in edf and csv files do not match")
    start_time_count:int = -1
    start_time_end:float = None
    results:list[tuple[str,float,float]] = []
    for label,sender,res, time_end,time_run in zip(labels,senders,responses,time_ends,time_runs):
        if res == sync_marker_name:
            start_time_count += 1
            start_time_end = time_ends[sync_lab_annos_indexes[start_time_count]]
        if sender != marker_name:
            continue
        offset_time_run = ((time_run - start_time_end) / 1000.0) + sync_edf_annos[start_time_count][1]
        offset_time_end = ((time_end - start_time_end) / 1000.0) + sync_edf_annos[start_time_count][1]
        results.append((label,offset_time_run,offset_time_end))

    def copied_func(_ ,wedf:pyedflib.EdfWriter,signals:list[ndarray]):
        for l,otr,ote, in results:
            mn = marker_name
            if not(l is None):
                mn += f"_{l}"
            wedf.writeAnnotation(otr,-1,mn)
            if not (end_marker_name is None):
                wedf.writeAnnotation(ote + end_marker_offset,-1,end_marker_name)
        if preprocessing_func is not None:
            return preprocessing_func(signals)
        return signals
    edf.copy(edf_reader,export_path,copied_func)
